@@ -1,19 +1,27 @@
 # -*- coding: utf-8 -*-
 # Copyright (c) 2018-2021, earthobservations developers.
 # Distributed under the MIT License. See LICENSE for more info.
+import logging
+import operator
 from abc import abstractmethod
 from enum import Enum
-from typing import Dict, Generator, Tuple, Union
+from typing import Dict, Generator, List, Tuple, Union
 
+import numpy as np
 import pandas as pd
+from pint import Quantity
 from pytz import timezone
 from tqdm import tqdm
 
 from wetterdienst.core.scalar.result import StationsResult, ValuesResult
 from wetterdienst.metadata.columns import Columns
-from wetterdienst.metadata.resolution import Resolution, ResolutionType
+from wetterdienst.metadata.resolution import Resolution
 from wetterdienst.metadata.timezone import Timezone
+from wetterdienst.metadata.unit import REGISTRY, OriginUnit, SIUnit
 from wetterdienst.util.enumeration import parse_enumeration_from_template
+from wetterdienst.util.logging import TqdmToLogger
+
+log = logging.getLogger(__name__)
 
 
 class ScalarValuesCore:
@@ -21,23 +29,41 @@ class ScalarValuesCore:
 
     # Fields for type coercion, needed for separation from fields with actual data
     # that have to be parsed differently when having data in tabular form
-    _meta_fields = [
-        Columns.STATION_ID.value,
-        Columns.DATE.value,
-        Columns.PARAMETER.value,
-        Columns.QUALITY.value,
-    ]
+    @property
+    def _meta_fields(self) -> List[str]:
+        """
+        Metadata fields that are independent of actual values and should be parsed
+        differently
+
+        :return: list of strings representing the metadata fields/columns
+        """
+        if not self.stations.stations.tidy:
+            fields = [
+                Columns.STATION_ID.value,
+                Columns.DATE.value,
+            ]
+        else:
+            fields = [
+                Columns.STATION_ID.value,
+                Columns.DATASET.value,
+                Columns.PARAMETER.value,
+                Columns.DATE.value,
+                Columns.VALUE.value,
+                Columns.QUALITY.value,
+            ]
+
+        return fields
 
     # Fields for date coercion
     _date_fields = [Columns.DATE.value, Columns.FROM_DATE.value, Columns.TO_DATE.value]
 
     # TODO: add data type (forecast, observation, ...)
-    @property
-    @abstractmethod
-    def _has_quality(self) -> bool:
-        """Attribute that tells if a weather service has quality, which otherwise will
-        have to be set to NaN"""
-        pass
+    # @property
+    # @abstractmethod
+    # def _has_quality(self) -> bool:
+    #     """Attribute that tells if a weather service has quality, which otherwise will
+    #     have to be set to NaN"""
+    #     pass
 
     @property
     def data_tz(self) -> timezone:
@@ -71,6 +97,13 @@ class ScalarValuesCore:
 
     @property
     def _complete_dates(self) -> pd.DatetimeIndex:
+        """
+        Complete datetime index for the requested start and end date, used for
+        building a complementary pandas DataFrame with the date column on which
+        other DataFrames can be joined on
+
+        :return: pandas.DatetimeIndex
+        """
         start_date, end_date = self.stations.start_date, self.stations.end_date
 
         if self.stations.stations.resolution == Resolution.MONTHLY:
@@ -89,13 +122,108 @@ class ScalarValuesCore:
 
     @property
     def _base_df(self) -> pd.DataFrame:
-        """Base dataframe which is used for creating empty dataframes if no data is
-        found or for merging other dataframes on the full dates"""
+        """
+        Base dataframe which is used for creating empty dataframes if no data is
+        found or for merging other dataframes on the full dates
+
+        :return: pandas DataFrame with a date column with complete dates
+        """
         return pd.DataFrame({Columns.DATE.value: self._complete_dates})
 
-    @staticmethod
-    def _determine_resolution(series: pd.Series) -> Resolution:
-        pass
+    def convert_values_to_si(self, df: pd.DataFrame, dataset) -> pd.DataFrame:
+        """
+        Function to convert values to metric units with help of conversion factors
+
+        :param df: pandas DataFrame that should be converted to SI units
+        :param dataset: dataset for which the conversion factors are created
+        :return: pandas DataFrame with converted (SI) values
+        """
+
+        def _convert_values_to_si(series):
+            """
+            Helper function to apply conversion factors column wise to a pandas DataFrame
+
+            :param series: pandas Series that should be converted
+            :return: converted pandas Series
+            """
+            op, factor = conversion_factors.get(series.name, (None, None))
+
+            if not op or not factor:
+                return series
+
+            return op(series, factor)
+
+        conversion_factors = self._create_conversion_factors(dataset)
+
+        df = df.apply(_convert_values_to_si, axis=0)
+
+        return df
+
+    def _create_conversion_factors(
+        self, dataset
+    ) -> Dict[str, Tuple[Union[operator.add, operator.mul], float]]:
+        """
+        Function to create conversion factors based on a given dataset
+
+        :param dataset: dataset for which conversion factors are created
+        :return: dictionary with conversion factors for given parameter name
+        """
+        dataset = dataset.name
+
+        dataset_accessor = self.stations.stations._dataset_accessor
+
+        if self.stations.stations._unique_dataset:
+            units = self.stations.stations._unit_tree[dataset_accessor]
+        else:
+            units = self.stations.stations._unit_tree[dataset_accessor][dataset]
+
+        conversion_factors = {}
+
+        # TODO eventually we may split this into smaller functions
+        for parameter in units:
+            origin_unit, si_unit = parameter.value
+
+            # Get parameter name
+            parameter = parameter.name
+
+            if self.stations.stations._unique_dataset:
+                parameter_value = self.stations.stations._dataset_tree[
+                    dataset_accessor
+                ][parameter].value
+            else:
+                parameter_value = self.stations.stations._dataset_tree[
+                    dataset_accessor
+                ][dataset][parameter].value
+
+            if si_unit == SIUnit.KILOGRAM_PER_SQUARE_METER.value:
+                # Fixed conversion factors to kg / m², as it only applies
+                # for water with density 1 g / cm³
+                if origin_unit == OriginUnit.MILLIMETER.value:
+                    conversion_factors[parameter_value] = (operator.mul, 1)
+                else:
+                    raise ValueError(
+                        "manually set conversion factor for precipitation unit"
+                    )
+            elif si_unit == SIUnit.DEGREE_KELVIN.value:
+                # Apply offset addition to temperature measurements
+                # Take 0 as this is appropriate for adding on other numbers
+                # (just the difference)
+                degree_offset = Quantity(0, origin_unit).to(si_unit).magnitude
+
+                conversion_factors[parameter_value] = (operator.add, degree_offset)
+            elif si_unit == SIUnit.PERCENT.value:
+                factor = REGISTRY(str(origin_unit)).to(str(si_unit)).magnitude
+
+                conversion_factors[parameter_value] = (operator.mul, factor)
+            else:
+                # For multiplicative units we need to use 1 as quantity to apply the
+                # appropriate factor
+                conversion_factors[parameter_value] = (
+                    operator.mul,
+                    Quantity(1, origin_unit).to(si_unit).magnitude,
+                )
+
+        return conversion_factors
 
     def __init__(self, stations: StationsResult) -> None:
         self.stations = stations
@@ -139,17 +267,18 @@ class ScalarValuesCore:
         )
         pass
 
-    def _get_empty_station_parameter_df(
-        self, station_id: str, parameter: Union[Enum, Tuple[Enum, Enum]]
+    def _create_empty_station_parameter_df(
+        self, station_id: str, parameter: Enum, dataset: Enum
     ) -> pd.DataFrame:
+        """
+        Function to create an empty DataFrame
+        :param station_id:
+        :param parameter:
+        :return:
+        """
         dataset_tree = self.stations.stations._dataset_tree
         resolution = self.stations.stations.resolution
 
-        dataset = None
-        # Split parameter into parameter and dataset if parameters are stored in dataset
-        if self.stations.stations._has_datasets:
-            parameter, dataset = parameter
-
         # if parameter is a whole dataset, take every parameter from the dataset instead
         if parameter == dataset:
             if self.stations.stations._unique_dataset:
@@ -158,6 +287,9 @@ class ScalarValuesCore:
                 parameter = [*dataset_tree[resolution.name][dataset.name]]
 
         if self.stations.stations.tidy:
+            if not self.stations.stations.start_date:
+                return pd.DataFrame(None, columns=self._meta_fields)
+
             data = []
             for par in pd.Series(parameter):
                 if par.name.startswith("QUALITY"):
@@ -177,11 +309,15 @@ class ScalarValuesCore:
 
             return df
         else:
-            df = self._base_df
             parameter = pd.Series(parameter).map(lambda x: x.value).tolist()
 
             # Base columns
-            columns = [Columns.STATION_ID.value, Columns.DATE.value, *parameter]
+            columns = [*self._meta_fields, *parameter]
+
+            if self.stations.stations.start_date:
+                return pd.DataFrame(None, columns=columns)
+
+            df = self._base_df
 
             df = df.reindex(columns=columns)
 
@@ -190,7 +326,7 @@ class ScalarValuesCore:
             return df
 
     def _build_complete_df(
-        self, df: pd.DataFrame, station_id: str, parameter: Enum
+        self, df: pd.DataFrame, station_id: str, parameter: Enum, dataset: Enum
     ) -> pd.DataFrame:
         # For cases where requests are not defined by start and end date but rather by
         # periods, use the returned df without modifications
@@ -198,10 +334,6 @@ class ScalarValuesCore:
         if not self.stations.start_date:
             return df
 
-        dataset = None
-        if self.stations.stations._has_datasets:
-            parameter, dataset = parameter
-
         if parameter != dataset or not self.stations.stations.tidy:
             df = pd.merge(
                 left=self._base_df,
@@ -263,58 +395,85 @@ class ScalarValuesCore:
 
             return pd.concat(data)
 
+    def _organize_df_columns(self, df: pd.DataFrame) -> pd.DataFrame:
+        """
+        Method to reorder index to always have the same order of columns
+
+        :param df:
+        :return:
+        """
+        columns = self._meta_fields
+
+        columns.extend(df.columns.difference(columns, sort=False))
+
+        df = df.reindex(columns=columns)
+
+        return df
+
     def query(self) -> Generator[ValuesResult, None, None]:
-        """Core method for data collection, iterating of station ids and yielding a
+        """
+        Core method for data collection, iterating of station ids and yielding a
         DataFrame for each station with all found parameters. Takes care of type
-        coercion of data, date filtering and humanizing of parameters."""
+        coercion of data, date filtering and humanizing of parameters.
+
+        :return:
+        """
         for station_id in self.stations.station_id:
             # TODO: add method to return empty result with correct response string e.g.
             #  station id not available
             station_data = []
 
-            for parameter in self.stations.parameter:
-                parameter_df = self._collect_station_parameter(station_id, parameter)
+            for parameter, dataset in self.stations.parameter:
+                parameter_df = self._collect_station_parameter(
+                    station_id, parameter, dataset
+                )
 
-                # TODO: solve exceptional case where empty df and dynamic resolution
-                if self.stations._resolution_type == ResolutionType.DYNAMIC:
-                    self.stations.resolution = self._determine_resolution(
-                        parameter_df[Columns.DATE.value]
-                    )
-
-                # TODO: move tidying of data outside collect method to apply it here
-                #  after collecting data/ if data is empty "tidy" the created base_df
-                #  instead
                 if parameter_df.empty:
-                    parameter_df = self._get_empty_station_parameter_df(
-                        station_id, parameter
-                    )
-                else:
-                    # Merge on full date range if values are found to ensure result
-                    # even if no actual values exist
-                    self._coerce_date_fields(parameter_df)
+                    continue
 
-                    parameter_df = self._build_complete_df(
-                        parameter_df, station_id, parameter
-                    )
+                # Merge on full date range if values are found to ensure result
+                # even if no actual values exist
+                self._coerce_date_fields(parameter_df)
+
+                parameter_df = self._coerce_parameter_types(parameter_df)
+
+                if self.stations.stations.si_units:
+                    parameter_df = self.convert_values_to_si(parameter_df, dataset)
+
+                if self.stations.stations.tidy:
+                    parameter_df = self.tidy_up_df(parameter_df, dataset)
+
+                    if parameter != dataset:
+                        parameter_df = parameter_df[
+                            parameter_df[Columns.PARAMETER.value]
+                            == parameter.value.lower()
+                        ]
+
+                parameter_df = self._build_complete_df(
+                    parameter_df, station_id, parameter, dataset
+                )
+
+                parameter_df = self._organize_df_columns(parameter_df)
 
                 station_data.append(parameter_df)
 
-            station_df = pd.concat(station_data, ignore_index=True)
+            try:
+                station_df = pd.concat(station_data, ignore_index=True)
+            except ValueError:
+                station_df = self._create_empty_station_parameter_df(
+                    station_id, parameter
+                )
 
             station_df = self._coerce_meta_fields(station_df)
-            station_df = self._coerce_parameter_types(station_df)
 
             # Filter for dates range if start_date and end_date are defined
-            if self.stations.start_date:
-                # df_station may be empty depending on if station has data for given
-                # constraints
-                try:
-                    station_df = station_df[
-                        (station_df[Columns.DATE.value] >= self.stations.start_date)
-                        & (station_df[Columns.DATE.value] <= self.stations.end_date)
-                    ]
-                except KeyError:
-                    pass
+            if not station_df.empty and self.stations.start_date:
+                station_df = station_df[
+                    (station_df[Columns.DATE.value] >= self.stations.start_date)
+                    & (station_df[Columns.DATE.value] <= self.stations.end_date)
+                ]
+
+            station_df = self._coerce_parameter_types(station_df)
 
             # Assign meaningful parameter names (humanized).
             if self.stations.humanize:
@@ -324,11 +483,13 @@ class ScalarValuesCore:
             if station_df.empty:
                 continue
 
-            # TODO: add meaningful metadata here
+            # TODO: add more meaningful metadata here
             yield ValuesResult(stations=self.stations, df=station_df)
 
     @abstractmethod
-    def _collect_station_parameter(self, station_id: str, parameter) -> pd.DataFrame:
+    def _collect_station_parameter(
+        self, station_id: str, parameter: Enum, dataset: Enum
+    ) -> pd.DataFrame:
         """
         Implementation of data collection for a station id plus parameter from the
         specified weather service. Takes care of the gathering of the data and putting
@@ -336,15 +497,56 @@ class ScalarValuesCore:
         of station id, date, parameter, value and quality in one row.
 
         :param station_id: station id for which the data is being collected
-        :param parameter: parameter for which the data is collected, parameter could
-        also be a tuple or a more complex object depending on the self._parse_parameters
-        function
+        :param parameter: parameter for which the data is collected
+        :param dataset: dataset for which the data is collected
         :return: pandas.DataFrame with the data for given station id and parameter
         """
         pass
 
+    def tidy_up_df(self, df: pd.DataFrame, dataset: Enum) -> pd.DataFrame:
+        """
+        Function to tidy a DataFrame
+
+        :param df:
+        :param dataset:
+        :return:
+        """
+        df = self._tidy_up_df(df, dataset)
+
+        df[Columns.DATASET.value] = pd.Series(
+            dataset.name.lower(), index=df.index, dtype=str
+        )
+
+        df[Columns.VALUE.value] = pd.to_numeric(df[Columns.VALUE.value]).astype(float)
+
+        if Columns.QUALITY.value not in df:
+            df[Columns.QUALITY.value] = np.nan
+
+        df[Columns.QUALITY.value] = pd.to_numeric(df[Columns.QUALITY.value]).astype(
+            float
+        )
+
+        df.loc[df[Columns.VALUE.value].isna(), Columns.QUALITY.value] = np.NaN
+
+        return df
+
+    @abstractmethod
+    def _tidy_up_df(self, df: pd.DataFrame, dataset) -> pd.DataFrame:
+        """
+        Abstract method to be implemented by services to tidy a DataFrame
+
+        :param df:
+        :return:
+        """
+        pass
+
     def _coerce_date_fields(self, df: pd.DataFrame) -> pd.DataFrame:
-        """ Function for coercion of possible date fields """
+        """
+        Function for coercion of possible date fields
+
+        :param df:
+        :return:
+        """
         for column in (
             Columns.DATE.value,
             Columns.FROM_DATE.value,
@@ -372,50 +574,86 @@ class ScalarValuesCore:
             df[Columns.STATION_ID.value]
         ).astype("category")
 
-        if self.stations.tidy:
-            df[Columns.PARAMETER.value] = self._coerce_strings(
-                df[Columns.PARAMETER.value]
-            ).astype("category")
+        if self.stations.stations.tidy:
+            for column in (Columns.DATASET.value, Columns.PARAMETER.value):
+                df[column] = self._coerce_strings(df[column]).astype("category")
 
-            if self._has_quality:
-                df[Columns.QUALITY.value] = self._coerce_integers(
-                    df[Columns.QUALITY.value]
-                ).astype("category")
+            df[Columns.VALUE.value] = pd.to_numeric(df[Columns.VALUE.value]).astype(
+                float
+            )
+
+            df[Columns.QUALITY.value] = pd.to_numeric(df[Columns.QUALITY.value]).astype(
+                float
+            )
 
         return df
 
     def _parse_station_id(self, series: pd.Series) -> pd.Series:
-        """Dedicated method for parsing station ids, by default uses the same method as
-        parse_strings but could be modified by the implementation class"""
+        """
+        Dedicated method for parsing station ids, by default uses the same method as
+        parse_strings but could be modified by the implementation class
+
+        :param series:
+        :return:
+        """
         return self.stations.stations._parse_station_id(series)
 
     def _coerce_dates(self, series: pd.Series) -> pd.Series:
-        """Method to parse dates in the pandas.DataFrame. Leverages the data timezone
-        attribute to ensure correct comparison of dates."""
+        """
+        Method to parse dates in the pandas.DataFrame. Leverages the data timezone
+        attribute to ensure correct comparison of dates.
+
+        :param series:
+        :return:
+        """
         return pd.to_datetime(series, infer_datetime_format=True).dt.tz_localize(
             self.data_tz
         )
 
     @staticmethod
     def _coerce_integers(series: pd.Series) -> pd.Series:
-        """Method to parse integers for type coercion. Uses pandas.Int64Dtype() to
-        allow missing values."""
-        return pd.to_numeric(series, errors="coerce").astype(pd.Int64Dtype())
+        """
+        Method to parse integers for type coercion. Uses pandas.Int64Dtype() to
+        allow missing values.
+
+        :param series:
+        :return:
+        """
+        return (
+            pd.to_numeric(series, errors="coerce")
+            .astype(pd.Float64Dtype())
+            .astype(pd.Int64Dtype())
+        )
 
     @staticmethod
     def _coerce_strings(series: pd.Series) -> pd.Series:
-        """ Method to parse strings for type coercion. """
+        """
+        Method to parse strings for type coercion.
+
+        :param series:
+        :return:
+        """
         return series.astype(pd.StringDtype())
 
     @staticmethod
     def _coerce_floats(series: pd.Series) -> pd.Series:
-        """ Method to parse floats for type coercion. """
+        """
+        Method to parse floats for type coercion.
+
+        :param series:
+        :return:
+        """
         return pd.to_numeric(series, errors="coerce")
 
     def _coerce_irregular_parameter(self, series: pd.Series) -> pd.Series:
-        """Method to parse irregular parameters. This will raise an error if an
+        """
+        Method to parse irregular parameters. This will raise an error if an
         implementation has defined irregular parameters but has not implemented its own
-        method of parsing irregular parameters."""
+        method of parsing irregular parameters.
+
+        :param series:
+        :return:
+        """
         if self._irregular_parameters:
             raise NotImplementedError(
                 "implement _parse_irregular_parameter "
@@ -425,54 +663,41 @@ class ScalarValuesCore:
         return pd.Series(series)
 
     def _coerce_parameter_types(self, df: pd.DataFrame) -> pd.DataFrame:
-        """ Method for parameter type coercion. Depending on the shape of the data. """
-        if not self.stations.tidy:
-            for column in df.columns:
-                if column in self._meta_fields or column in self._date_fields:
-                    continue
-                if column in self._irregular_parameters:
-                    df[column] = self._coerce_irregular_parameter(df[column])
-                elif column in self._integer_parameters or column.startswith(
-                    Columns.QUALITY_PREFIX.value
-                ):
-                    df[column] = self._coerce_integers(df[column])
-                elif column in self._string_parameters:
-                    df[column] = self._coerce_strings(df[column])
-                else:
-                    df[column] = self._coerce_floats(df[column])
+        """
+        Method for parameter type coercion. Depending on the shape of the data.
 
-            return df
-
-        data = []
-        for parameter, group in df.groupby(Columns.PARAMETER.value, sort=False):
-            if parameter in self._irregular_parameters:
-                group[Columns.VALUE.value] = self._coerce_irregular_parameter(
-                    group[Columns.VALUE.value]
-                )
-            elif parameter in self._integer_parameters:
-                group[Columns.VALUE.value] = self._coerce_integers(
-                    group[Columns.VALUE.value]
-                )
-            elif parameter in self._string_parameters:
-                group[Columns.VALUE.value] = self._coerce_strings(
-                    group[Columns.VALUE.value]
-                )
+        :param df:
+        :return:
+        """
+        for column in df.columns:
+            if column in self._meta_fields or column in self._date_fields:
+                continue
+            if column in self._irregular_parameters:
+                df[column] = self._coerce_irregular_parameter(df[column])
+            elif column in self._integer_parameters or column.startswith(
+                Columns.QUALITY_PREFIX.value
+            ):
+                df[column] = self._coerce_integers(df[column])
+            elif column in self._string_parameters:
+                df[column] = self._coerce_strings(df[column])
             else:
-                group[Columns.VALUE.value] = self._coerce_floats(
-                    group[Columns.VALUE.value]
-                )
-
-            data.append(group)
-
-        df = pd.concat(data)
+                df[column] = self._coerce_floats(df[column])
 
         return df
 
     def all(self) -> ValuesResult:
-        """ Collect all data from self.collect_data """
+        """
+        Collect all data from self.query
+
+        :return:
+        """
         data = []
 
-        for result in tqdm(self.query(), total=len(self.stations.station_id)):
+        tqdm_out = TqdmToLogger(log, level=logging.INFO)
+
+        for result in tqdm(
+            self.query(), total=len(self.stations.station_id), file=tqdm_out
+        ):
             data.append(result.df)
 
         if not data:
@@ -481,22 +706,17 @@ class ScalarValuesCore:
         df = pd.concat(data, ignore_index=True)
 
         # Have to reapply category dtype after concatenation
-        for column in (
-            Columns.STATION_ID.value,
-            Columns.PARAMETER.value,
-            Columns.QUALITY.value,
-        ):
-            try:
-                df[column] = df[column].astype("category")
-            except KeyError:
-                pass
-
-        df.attrs["tidy"] = self.stations.tidy
+        df = self._coerce_meta_fields(df)
 
         return ValuesResult(stations=self.stations, df=df)
 
     def _humanize(self, df: pd.DataFrame) -> pd.DataFrame:
-        """ Method for humanizing parameters. """
+        """
+        Method for humanizing parameters.
+
+        :param df:
+        :return:
+        """
         hcnm = self._create_humanized_parameters_mapping()
 
         if not self.stations.tidy:
@@ -509,10 +729,14 @@ class ScalarValuesCore:
         return df
 
     def _create_humanized_parameters_mapping(self) -> Dict[str, str]:
-        """Method for creation of parameter name mappings based on
-        self._parameter_base"""
+        """
+        Method for creation of parameter name mappings based on
+        self._parameter_base
+
+        :return:
+        """
         hcnm = {
-            parameter.value: parameter.name
+            parameter.value: parameter.name.lower()
             for parameter in self.stations.stations._parameter_base
         }
 
